# test fixtures

import pytest


@pytest.fixture
def explorer():
    return "Etherscan"


@pytest.fixture
def chain():
    return "goerli"


@pytest.fixture
def address():
    return "0x6e0e0802A69522533988ae0AD3A7E9ebD3B2c05D"


@pytest.fixture
def test_event_name():
    return "RoleGranted"


@pytest.fixture
def expected_topic():
    return "0x2f8788117e7eff1d82e926ec794901d17c78024a50270940304540a733656f0d"


@pytest.fixture
def expected_functions():
    return [
        "DEFAULT_ADMIN_ROLE",
        "MINTER_ROLE",
        "PAUSER_ROLE",
        "approve",
        "authorizeEthersieve",
        "balanceOf",
        "burn",
        "contractURI",
        "getApproved",
        "getRoleAdmin",
        "grantRole",
        "hasRole",
        "isApprovedForAll",
        "name",
        "ownerOf",
        "pause",
        "paused",
        "renounceRole",
        "revokeRole",
        "safeMint",
        "safeTransferFrom",
        "setApprovalForAll",
        "setTokenURI",
        "supportsInterface",
        "symbol",
        "tokenURI",
        "transferFrom",
        "unpause",
    ]


@pytest.fixture
def expected_events():
    return [
        "Approval",
        "ApprovalForAll",
        "Paused",
        "RoleAdminChanged",
        "RoleGranted",
        "RoleRevoked",
        "Transfer",
        "Unpaused",
    ]



# abi module test

import os

import pytest

from eth_abi_tool import ABI


@pytest.fixture
def abi(explorer, chain, address):
    key = os.environ["EXPLORER_API_KEY"]
    abi = ABI(explorer, chain, key)
    abi.get(address)
    return abi


def test_abi_get(abi, address):
    contract_abi = abi.get(address)
    assert isinstance(contract_abi, list)
    for a in contract_abi:
        assert isinstance(a, dict)


def test_abi_functions(abi, expected_functions):
    functions = abi.functions()
    assert isinstance(functions, dict)
    assert set(functions.keys()) == set(expected_functions)
    for name, function_abi in functions.items():
        assert isinstance(name, str)
        assert isinstance(function_abi, dict)
        assert function_abi["type"] == "function"


def test_abi_function_names(abi, expected_functions):
    names = abi.function_names()
    assert set(names) == set(expected_functions)


def test_abi_events(abi, expected_events):
    events = abi.events()
    assert isinstance(events, dict)
    assert set(events.keys()) == set(expected_events)
    for name, event in events.items():
        assert isinstance(name, str)
        assert isinstance(event, dict)
        assert event["type"] == "event"


def test_abi_event_names(abi, expected_events):
    names = abi.event_names()
    assert set(names) == set(expected_events)


def test_abi_topic_by_name(abi, test_event_name, expected_topic):
    topic = abi.topic(name=test_event_name)
    assert topic == expected_topic


def test_abi_topic_by_event_abi(abi, test_event_name, expected_topic):
    event_abi = abi.event(test_event_name)
    topic = ABI().topic(event=event_abi)
    assert topic == expected_topic



"""ABI object"""

import json

import requests
from ratelimiter import RateLimiter

# rate limit to 5 calls per second
RATELIMIT = 5


class ABIException(Exception):
    pass


class ABIQueryFailure(ABIException):
    """Explorer API query failed"""

    pass


class ABIError(ABIException):
    """Explorer API returned error"""

    pass


class ABI:

    EXPLORERS = {
        "Etherscan": {
            "mainnet": {"url": "https://api.etherscan.io/api"},
            "kovan": {"url": "https://api-kovan.etherscan.io/api"},
            "goerli": {"url": "https://api-goerli.etherscan.io/api"},
            "rinkeby": {"url": "https://api-rinkeby.etherscan.io/api"},
            "ropsten": {"url": "https://api-ropsten.etherscan.io/api"},
            "sepolia": {"url": "https://api-sepolia.etherscan.io/api"},
        }
    }

    def __init__(self, explorer=None, chain=None, key=None):
        self.explorer = explorer or list(self.EXPLORERS.keys())[0]
        self.chain = chain or list(self.EXPLORERS[self.explorer].keys())[0]
        self.url = self.EXPLORERS[self.explorer][self.chain]["url"]
        self.key = key
        self.abi = None

    @RateLimiter(max_calls=RATELIMIT, period=1)
    def get(self, contract_address, key=None):
        from eth_utils import to_normalized_address

        key = key or self.key
        address = to_normalized_address(contract_address)
        params = dict(
            module="contract", action="getabi", address=address, apikey=key
        )
        response = requests.get(self.url, params=params)
        if not response:
            raise ABIQueryFailure(f"{response}")
        result = response.json()
        if result["status"] != "1" or result["message"] != "OK":
            raise ABIError(f"{result}")
        else:
            self.abi = json.loads(result["result"])
        return self.abi

    def _select_elements(self, _type):
        ret = {}
        for i in self.abi:
            if i.get("type", None) == _type:
                name = i.get("name", None)
                if name:
                    ret[name] = i
        return ret

    def functions(self):
        return self._select_elements("function")

    def function_names(self):
        return list(self.functions().keys())

    def function(self, name):
        return self.functions()[name]

    def events(self):
        return self._select_elements("event")

    def event_names(self):
        return list(self.events().keys())

    def event(self, event):
        return self.events()[event]

    def topic(self, *, name=None, event=None):
        """generate event signature hash"""
        from eth_utils import event_abi_to_log_topic, to_hex

        if name is not None:
            event = self.event(name)

        topic = event_abi_to_log_topic(event)
        ret = to_hex(topic)
        return ret



#!/usr/bin/env python

"""Tests for `eth_abi_tool` CLI"""

import json
from traceback import print_exception

import pytest
from click.testing import CliRunner

import eth_abi_tool
from eth_abi_tool import __version__, cli


def test_version():
    """Test reading version and module name"""
    assert eth_abi_tool.__name__ == "eth_abi_tool"
    assert __version__
    assert isinstance(__version__, str)


@pytest.fixture
def run():
    runner = CliRunner()

    # env = os.environ.copy()
    # env['EXTRA_ENV_VAR'] = 'VALUE'

    def _run(cmd, **kwargs):
        expect_exit_code = kwargs.pop("expect_exit_code", 0)
        expect_exception = kwargs.pop("expect_exception", None)
        # kwargs["env"] = env
        result = runner.invoke(cli, cmd, **kwargs)
        if result.exception:
            if expect_exception is None or not isinstance(
                result.exception, expect_exception
            ):
                print_exception(result.exception)
                assert False, result.exception
        else:
            assert result.exit_code == expect_exit_code, result.output
        return result

    return _run


def test_cli_none(run):
    result = run([], expect_exception=RuntimeError)
    assert "Usage" in result.output


def test_cli_help(run):
    result = run(["--help"])
    assert "Show this message and exit." in result.output


def test_cli_config_format_show(run):
    result = run(["config", "format", "show"])
    assert len(result.output)


def test_cli_config_format_set_none(run):
    result = run(["config", "format", "set"])
    assert result.exit_code == 0


def test_cli_config_format_set_fmt(run):
    result = run(["--fmt", "text", "config", "format", "set"])
    assert result.exit_code == 0


def test_cli_config_format_set_tablefmt(run):
    result = run(["--tablefmt", "plain", "config", "format", "set"])
    assert result.exit_code == 0


def test_cli_config_format_set_both(run):
    result = run(
        [
            "--fmt",
            "text",
            "--tablefmt",
            "fancy_grid",
            "config",
            "format",
            "set",
        ]
    )
    assert result.exit_code == 0


def test_cli_config_format_reset(run):
    result = run(["config", "format", "reset"])
    assert result.exit_code == 0


def test_cli_get_abi(run, address):
    result = run(["get", address])
    abi = json.loads(result.output)
    assert isinstance(abi, list)
    for a in abi:
        assert isinstance(a, dict)



"""Top-level package for eth-abi-tool."""

from .abi import ABI
from .cli import cli
from .version import __author__, __email__, __timestamp__, __version__

__all__ = ["cli", "ABI", __version__, __timestamp__, __author__, __email__]



"""Console script for eth_abi_tool."""


import json
import sys
from pathlib import Path

import click
import tablib
import tabulate
from box import Box

from .abi import ABI
from .exception_handler import ExceptionHandler
from .version import __timestamp__, __version__

header = f"{__name__.split('.')[0]} v{__version__} {__timestamp__}"

OUTPUT_FORMATS = ["json", "yaml", "csv", "cli", "text"]
DEFAULT_FMT = "json"
TABLE_FORMATS = tabulate._table_formats
DEFAULT_TABLEFMT = "pretty"

EXPLORERS = list(ABI.EXPLORERS.keys())
CHAINS = list(ABI.EXPLORERS[EXPLORERS[0]].keys())


def config_read(ctx):
    if ctx.obj.config_file.is_file():
        ctx.obj.config = Box(json.loads(ctx.obj.config_file.read_text()))
    else:
        ctx.obj.config = Box({})
    ctx.obj.config.setdefault("contracts", {})
    ctx.obj.config.setdefault("format", {})
    return ctx.obj.config


def config_write(ctx):
    ctx.obj.config_file.write_text(json.dumps(dict(ctx.obj.config), indent=2))


def get_default_format(ctx, key):
    return ctx.obj.config.format.get(key, DEFAULT_FMT)


def get_format(ctx, key):
    return ctx.obj.get(key) or get_default_format(ctx, key)


def output_table(ctx, data):
    fmt = get_format(ctx, "fmt")
    if fmt in ["cli", "text"]:
        tablefmt = get_format(ctx, "tablefmt")
        output = data.export("cli", tablefmt=tablefmt)
    elif fmt == "json":
        output = data.export(fmt)
        if not ctx.obj.compact:
            output = json.dumps(json.loads(output), indent=2)
    else:
        raise ValueError(f"unexpected {fmt=}")
    click.echo(output)


@click.group("abi")
@click.version_option(message=header)
@click.option("-d", "--debug", is_flag=True, help="debug mode")
@click.option(
    "-k",
    "--key",
    type=str,
    envvar="EXPLORER_API_KEY",
    show_envvar=True,
    help="Explorer API Key",
)
@click.option(
    "-C",
    "--chain",
    type=click.Choice(CHAINS),
    envvar="EXPLORER_CHAIN",
    show_envvar=True,
    help="Chain Name",
)
@click.option(
    "-e",
    "--explorer",
    type=click.Choice(EXPLORERS),
    envvar="EXPLORER_GATEWAY",
    show_envvar=True,
    help="Blockchain Explorer Name",
)
@click.option(
    "-c",
    "--config-file",
    type=click.Path(dir_okay=False, writable=True, path_type=Path),
    default=Path.home() / ".eth-abi-tool",
    help="config file",
)
@click.option(
    "-f",
    "--fmt",
    type=click.Choice(OUTPUT_FORMATS),
    default=None,
    help="output format",
)
@click.option(
    "-T",
    "--tablefmt",
    type=click.Choice(TABLE_FORMATS),
    default=None,
    help="text table output format",
)
@click.option(
    "-J/-j",
    "--compact/--no-compact",
    is_flag=True,
    default=None,
    help="compact json format",
)
@click.pass_context
def cli(ctx, **kwargs):
    """Ethereum Contract ABI Utility"""
    ctx.obj = Box(kwargs)
    ctx.obj.ehandler = ExceptionHandler(ctx.obj.debug)
    ctx.obj.abi = ABI(ctx.obj.explorer, ctx.obj.chain, ctx.obj.key)
    if ctx.obj.compact is not None:
        ctx.obj.fmt = "json"
    if ctx.obj.fmt == "text":
        ctx.obj.fmt = "cli"
    config_read(ctx)


@cli.group
@click.pass_context
def config(ctx):
    """config commands"""
    pass


@config.group
@click.pass_context
def format(ctx):
    """output format config"""
    pass


@format.command(name="reset")
@click.pass_context
def config_format_reset(ctx):
    """reset default output format"""
    ctx.obj.config.format.fmt = DEFAULT_FMT
    ctx.obj.config.format.tablefmt = DEFAULT_TABLEFMT
    config_write(ctx)


@format.command(name="set")
@click.pass_context
def config_format_set(ctx):
    """set current output format selections as default"""
    ctx.obj.config.format.fmt = ctx.obj.fmt or DEFAULT_FMT
    ctx.obj.config.format.tablefmt = ctx.obj.tablefmt or DEFAULT_TABLEFMT
    config_write(ctx)


@format.command(name="show")
@click.pass_context
def config_format_show(ctx):
    """display output format default settings"""
    data = tablib.Dataset()
    data.append(("fmt", get_default_format(ctx, "fmt")))
    data.append(("tablefmt", get_default_format(ctx, "tablefmt")))
    data.headers = ["setting", "value"]
    ctx.obj.fmt = "cli"
    ctx.obj.tablefmt = "pretty"
    output_table(ctx, data)


@config.group
@click.pass_context
def contract(ctx):
    """contract config"""
    pass


@contract.command(name="add")
@click.argument("name", type=str)
@click.argument("address", type=str)
@click.pass_context
def config_contract_add(ctx, name, address):
    """add contract name and address"""
    from eth_utils import is_address

    if not is_address(address):
        raise ValueError(f"{address=}")
    ctx.obj.config.contracts[name] = address
    config_write(ctx)


@contract.command(name="delete")
@click.argument("name", type=str)
@click.pass_context
def config_contract_delete(ctx, name):
    """delete named contract"""
    del ctx.obj.config.contracts[name]
    config_write(ctx)


@contract.command(name="list")
@click.pass_context
def config_contract_list(ctx):
    """list configured contract addresses"""
    from eth_utils import to_normalized_address

    contracts = ctx.obj.config.contracts
    data = tablib.Dataset()
    for name, address in contracts.items():
        data.append([name, to_normalized_address(address)])
    data.headers = ["Name", "Address"]
    output_table(ctx, data)


@cli.command
@click.option("-f", "--functions", is_flag=True, help="select only functions")
@click.option("-e", "--events", is_flag=True, help="select only events")
@click.option(
    "-h/-H",
    "--header/--no-header",
    is_flag=True,
    default=True,
    help="select header output",
)
@click.option("-n", "--name", is_flag=True, help="switch name output")
@click.option("-a", "--abi", is_flag=True, help="select abi output")
@click.option("-t", "--topic", is_flag=True, help="select topic0 hex output")
@click.option("-c", "--complete", is_flag=True, help="output complete ABI")
@click.argument("contract", type=str)
@click.argument("abi-name", type=str, required=False)
@click.pass_context
def get(
    ctx,
    functions,
    events,
    header,
    name,
    abi,
    topic,
    complete,
    contract,
    abi_name,
):
    """query block explorer and output contract ABI as json"""
    from eth_utils import is_address

    if not any([functions, events, header, name, abi, topic]):
        complete = True

    if functions and not any([events, header, name, abi, topic]):
        name = True

    if events and not any([functions, header, name, abi, topic]):
        name = True

    if contract in ctx.obj.config.contracts.keys():
        contract = ctx.obj.config.contracts[contract]
    elif not is_address(contract):
        raise ValueError(f"{contract=}")

    result = ctx.obj.abi.get(contract)

    if complete:
        if ctx.obj.compact:
            indent = None
        else:
            indent = 2
        click.echo(json.dumps(result, indent=indent))
    else:
        if functions:
            abi_type = "function"
        elif events:
            abi_type = "event"
        else:
            abi_type = None
        output_abi(ctx, result, abi_type, header, name, abi, topic, abi_name)


def output_abi(
    ctx,
    contract_abi,
    abi_type,
    enable_header,
    enable_name,
    enable_abi,
    enable_topic,
    name=None,
):
    """format abi output

    ctx.obj.header: ------- bool include header
    contract_abi: --------- list of abi elements
    abi_type: ------------- None / 'function' / 'event'
    enable_header: bool --- include header in output
    enable_name: bool ----- include name in output
    enable_abi: bool ------ include abi in output
    enable_topic: bool ---- include hex topic0 in output
    name: str ------------- include only matching rows if not None

    """

    data = tablib.Dataset()
    for abi in [Box(a) for a in contract_abi]:
        # skip rows with no name
        if "name" not in abi:
            continue
        # if name specified, skip non-matching rows
        if name is not None and abi.name != name:
            continue

        # if abi_type specified, skip non-matching rows
        if abi_type is not None and abi.type != abi_type:
            continue

        row = []
        header = []

        if enable_name:
            header.append("name")
            row.append(abi.name)
        if enable_abi:
            header.append("abi")
            row.append(abi.to_dict())
            # if ctx.obj.get('fmt') == 'json':
            #    row.append(abi.to_dict())
            # else
            #    row.append(compressed_json(abi))
        if enable_topic:
            header.append("topic")
            if abi.type == "event":
                topic = ABI().topic(event=abi)
            else:
                topic = ""
            row.append(topic)

        if row is not None and len(row) > 0:
            data.append(tuple(row))

        if len(header) > 0 and enable_header:
            data.headers = header
    output_table(ctx, data)


def compressed_json(abi):
    return json.dumps(abi.to_dict(), separators=[":", ","])


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover
